# Copyright Commonwealth of Australia, Bureau of Meteorology 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import pyearthtools.zoo


@pytest.mark.parametrize(
    "value, result",
    [
        ("test", ("test", None)),
        ("test{value=1}", ("test", {"value": 1})),
        ("test{value=1.0}", ("test", {"value": 1.0})),
        ("test{value=true}", ("test", {"value": True})),
    ],
)
def test_split_name_assignment(value, result):
    assert pyearthtools.zoo.utils.split_name_assignment(value) == result


@pytest.mark.parametrize(
    "value",
    [
        ("test{value=1}{wow=morevariables}"),
    ],
)
def test_split_name_assignment_error(value):
    with pytest.raises(ValueError):
        pyearthtools.zoo.utils.split_name_assignment(value)
