# Copyright Commonwealth of Australia, Bureau of Meteorology 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations


from typing import Any
from pathlib import Path

import pyearthtools.data
import pyearthtools.zoo
import pyearthtools.training
from numpy import ndarray

CONFIG_PATH = Path(__file__, "../configs/").resolve()


class FakeModelWrapper(pyearthtools.training.wrapper.ModelWrapper):
    def predict(self, data: Any, **kwargs) -> ndarray:
        return data

    ## Model State Functions
    def load(self, path: str | Path | bool, **kwargs):
        pass

    def save(self, path: str | Path, **kwargs):
        raise NotImplementedError()


@pyearthtools.zoo.register("Testing/FakeModel")
class FakeModel(pyearthtools.zoo.BaseForecastModel):
    _name = "Testing/FakeModel"
    _default_config_path = CONFIG_PATH

    def load(self, *args, **kwargs) -> tuple[Any, dict[str, Any]]:
        return pyearthtools.training.predict.TimeSeriesAutoRecurrentPredictor(FakeModelWrapper(None, self.pipeline)), {
            "data_interval": (1, "hour"),
            **kwargs,
        }
